package restlet

import (
    "net/http"
    "strings"
    "crypto/sha1"
    "fmt"
    "time"
    "errors"
    "io/ioutil"
    "encoding/json"
    log "github.com/Sirupsen/logrus"
    "cygnux/kepler/utils"
    "github.com/gorilla/mux"
)

var _debugInfo bool = false

func _get_and_split(h utils.HashMap, key string) []string {
    s, _ := h.GetString(key, "")
    return strings.Split(s, ",")
}

func Build_QueryControl(queries utils.HashMap) *QueryController {

    qc := &QueryController{}
    qc.Offset, _ = queries.GetUInt(QCTRL_OFFSET, QCTRL_OFFSET_DEFAULT)
    qc.Limit, _ = queries.GetUInt(QCTRL_LIMIT, QCTRL_LIMIT_DEFAULT)
    qc.Includes = _get_and_split(queries, QCTRL_INCLUDES)
    qc.Excludes = _get_and_split(queries, QCTRL_EXCLUDES)
    qc.OrderBy = _get_and_split(queries, QCTRL_ORDERBY)
    qc.GroupBy = _get_and_split(queries, QCTRL_GROUPBY)
    return qc
}

func MatchMethod(m string, methods []string) bool {
    if len(methods) == 0 || methods[0] == "*" {
        return true
    }
    for _, s := range methods {
        if s == m {
            return true
        }
    }
    return false
}

func HashKey(s ...string) string {
    h := sha1.New()
    ns := strings.Join(s, "")
    h.Write([]byte(ns))
    bs := h.Sum(nil)
    return fmt.Sprintf("%x", bs)
}

func SetDebug(d bool) {
    _debugInfo = d
}

func Failure_Response(code uint, message string) (*RestletResult, error) {
    return &RestletResult{
        Code: code,
        Message: message,
    }, errors.New(message)
}

func MakeRestletHandler(h RestletHandler, predictor RequestPredictor, ctx_provider RequestContextProvider, methods []string, cache *CacheController) http.Handler {
    var f = func(response http.ResponseWriter, request *http.Request) {
        var t1 int64
        if _debugInfo {
            t1 = time.Now().Unix()
        }
        log.Infoln(">", request.Method, request.URL.String())
        if msg, ret := predictor.Predicate(request); !ret {
            response.Header().Set("Content-Type", "text/plain")
            response.WriteHeader(403)
            response.Write([]byte(msg))
            return
        }
        if ! MatchMethod(request.Method, methods) {
            response.Header().Set("Content-Type", "text/plain")
            response.WriteHeader(405)
            response.Write([]byte(fmt.Sprintf("Method '%s' not allowed!", request.Method)))
            return
        }
        //var ctx *RequestContext
        var url_params utils.HashMap
        var queries utils.HashMap = utils.HashMap(make(map[string]string))
        var post_data []byte
        var c_key string
        ctx := ctx_provider.NewContext(request)
        url_params = utils.HashMap(mux.Vars(request))
        for k, v := range request.URL.Query() {
            queries[k] = v[0]
        }
        if nil != cache && MatchMethod(request.Method, cache.CacheMethods) {
            c_key = cache.KeyPrefix + HashKey(request.Method, request.URL.String())
            if d, t := ctx.Cache().Get(c_key); t {
                log.Debugln("Restlet:> Hit cache for :> ", c_key, request.Method, request.URL.String())
                content_type := string(d[:CONTENT_HEADER_SIZE])  // http.DetectContentType(d)
                log.Debugln("Restlet:> detect content-type:", content_type)
                response.Header().Set("Content-Type", content_type)
                response.WriteHeader(200)
                response.Write(d[CONTENT_HEADER_SIZE:])
                return
            }
        }
        switch request.Method {
        case "POST", "PUT", "PATCH":
            if buf, e := ioutil.ReadAll(request.Body); e == nil {
                post_data = buf
            } else {
                log.Errorln("Restlet:> Read Body error:>", e)
                response.Header().Set("Content-Type", "text/plain")
                response.WriteHeader(FATAL_INTERNAL_SERVER_ERROR)
                response.Write([]byte(fmt.Sprintf("Read Body error:> %s", e)))
                return
            }
        }
        if r, e := h.Handle(ctx, url_params, queries, post_data); nil != e {
            if nil == r {
                response.Header().Set("Content-Type", "text/plain")
                response.WriteHeader(FATAL_INTERNAL_SERVER_ERROR)
                response.Write([]byte(fmt.Sprintf("Failed:> %s", e)))
            } else {
                response.Header().Set("Content-Type", "text/plain")
                response.WriteHeader(int(r.Code / 100))
                response.Write([]byte(r.Message))
            }

        } else {
            r.ETag = fmt.Sprintf("%016x", time.Now().Unix())
            if _debugInfo {
                r.Debug = &DebugResult{Start:t1, Finish:time.Now().Unix()}
            }

            var output interface{}
            var err error
            var pbytes []byte
            var ok bool
            if r.DataOnly {
                output = r.Data
            } else {
                output = r
            }
            if r.RawBytes {
                err = nil
                pbytes, ok = output.([]byte)
                if !ok {
                    panic("Restlet:> Data not row bytes!")
                }
            } else {
                r.ContentType = "application/json"
                pbytes, err = json.Marshal(output)
            }
            if nil != err {
                response.Header().Set("Content-Type", "text/plain")
                response.WriteHeader(FATAL_INTERNAL_SERVER_ERROR/100)
                response.Write([]byte(fmt.Sprintf("Failed:> %s", e)))
            } else {
                statusCode := int(r.Code / 100)
                if statusCode / 100 < 1 || statusCode / 100 > 5 {
                    statusCode = 200
                }
                response.Header().Set("Content-Type", r.ContentType)
                response.WriteHeader(statusCode)
                response.Write(pbytes)
                if nil != cache && MatchMethod(request.Method, cache.CacheMethods) {
                    var ct_bytes [CONTENT_HEADER_SIZE]byte
                    copy(ct_bytes[:], []byte(r.ContentType)[:CONTENT_HEADER_SIZE])
                    ctx.Cache().Set(c_key, append(ct_bytes[:], pbytes...), cache.Expires)
                    log.Debugln("Restlet:> Cached Content for:> ", c_key, request.Method, request.URL.String())
                }
            }
        }
    }
    return http.HandlerFunc(f)
}