package restlet

import (
    "net/http"
    "database/sql"
    "cygnux/kepler/utils"
)

type ControlResult struct {
    Total  uint64  `json:"total"`
    Count  uint  `json:"count"`
    Offset uint `json:"offset"`
    Limit  uint  `json:"limit"`
}

type DebugResult struct {
    Start  int64
    Finish int64
}

type RestletResult struct {
    Code        uint `json:"code"`
    Message     string  `json:"message,omitempty"`
    ContentType string `json:"-"`
    DataOnly    bool `json:"-"`
    RawBytes    bool `json:"-"`
    Headers     http.Header `json:"-"`
    Cookies     []*http.Cookie `json:"-"`
    ETag        string `json:"__eTag,omitempty"`
    Debug       *DebugResult `json:"__debug,omitempty"`
    Model       string    `json:"__model,omitempty"`
    IsList      bool      `json:"__is_list,omitempty"`
    Control     *ControlResult  `json:"__control,omitempty"`
    Data        interface{}    `json:"data"`
}

type RawletResult struct {
    Code        uint
    Message     string
    ContentType string
    Data        []byte
}

type QueryController struct {
    Offset   uint
    Limit    uint
    Includes []string
    Excludes []string
    OrderBy  []string
    GroupBy  []string
    Filters  map[string]interface{}
}

type RequestCache interface {
    // Get Cached bytes via k, return empty bytes and false if failed.
    Get(k string) ([]byte, bool)
    // Set bytes to cache with given key and optional expires
    Set(k string, d []byte, x...uint)
}

type MessageQueue interface {
    Publish(channel string, body []byte) error
    Subscribe(channel string) ([]byte, error)
    Send(queue string, body []byte) error
    Recv(queue string) ([]byte, error)
}

type RequestContext interface {
    Schema() string
    SQL() *sql.DB
    Cache() RequestCache
    Queue() MessageQueue
    Chan() chan TaskObject
    Option() interface{}
    Request() *http.Request
}

type RequestContextProvider interface {
    NewContext(*http.Request) RequestContext
}

type RequestPredictor interface {
    Predicate(*http.Request) (string, bool)
}

type CacheController struct {
    KeyPrefix    string
    CacheMethods []string
    CacheStatus  []int
    Expires      uint
}

type RestletHandler interface {
    Handle(ctx RequestContext, url_params map[string]string, queries map[string]string, post_data []byte) (*RestletResult, error)
}

type RawletHandler interface {
    Handle(ctx RequestContext, w http.ResponseWriter, r *http.Request)
}

type TaskContext interface {
    Schema() string
    SQL() *sql.DB
    Cache() RequestCache
    Queue() MessageQueue
    Chan() chan TaskObject
    Option() interface{}
}

type TaskContextProvider interface {
    NewContext() TaskContext
}

type TaskObject struct {
    Queue  string
    Params []interface{}
}

type TaskletHandler interface {
    Handle(TaskContext, ...interface{}) error
}

type RestletHandleFunc func(ctx RequestContext, url_params utils.HashMap, queries utils.HashMap, post_data []byte) (*RestletResult, error)

type TaskletHandlerFunc func(TaskContext, ...interface{}) error

func (f RestletHandleFunc) Handle(ctx RequestContext, url_params map[string]string, queries map[string]string, post_data []byte) (*RestletResult, error) {
    return f(ctx, url_params, queries, post_data)
}

func (f TaskletHandlerFunc) Handle(ctx TaskContext, params ...interface{}) error {
    return f(ctx, params...)
}